# -*- coding: utf-8 -*-
r"""
Generate Org-Mode agenda file from ical web sources
===================================================

Goal of the script
"""
# Created: Fri Jan 11 22:54:35 2019
# Author: Óscar Nájera
# License:GPL-3

import argparse
import asyncio
import configparser
import hashlib
import logging
import os

import requests
import aiohttp

from org_agenda.ical2org import org_events
from org_agenda.cards2org import org_contacts

LOGGER = logging.getLogger("Org_calendar")
LOGGER.addHandler(logging.StreamHandler())


async def passwordstore(address: str) -> str:
    """Get password from passwordstore address"""

    process = await asyncio.create_subprocess_shell(
        f"pass show {address}", stdout=asyncio.subprocess.PIPE
    )
    stdout, _ = await process.communicate()

    if process.returncode == 0:
        return stdout.decode("utf-8").split()[0]

    raise ValueError("Unknown password address")


async def download(url, section, force, session):
    "Return calendar from download or cache"
    hash_url = hashlib.sha1(url.encode("UTF-8")).hexdigest()[:10]
    url_file = f"/tmp/agenda-{hash_url}"
    if not force and os.path.exists(url_file):
        with open(url_file) as txt:
            return txt.read()

    user = section["user"]
    password = await passwordstore(section["passwordstore"])

    LOGGER.info("Downloading from: %s", url)
    async with session.get(url, auth=aiohttp.BasicAuth(user, password)) as reply:
        if reply.status == 200:
            text = await reply.text()
            with open(url_file, "w") as txt:
                txt.write(text)
            return text

        raise ValueError(f"Could not get calendar: {url}\n{reply}")


async def get_resource(config, resource, force):
    "RESOURCE{calendars,addressbooks} are returned given the CONFIG's urls, FORCE download"

    cal = []
    async with aiohttp.ClientSession() as session:
        for section in config.sections():
            for entry in config[section].get(resource, "").split():
                url = config[section]["url"].format(entry)
                cal.append(download(url, config[section], force, session))

        return await asyncio.gather(*cal)


def write_agenda(config, args):
    "Write the agenda to file"

    if args.url:
        calendars = [requests.get(args.url, auth=("username", "")).text]
    else:
        calendars = asyncio.run(get_resource(config, "calendars", args.force))

    ahead = int(config["DEFAULT"].get("ahead", 50))
    back = int(config["DEFAULT"].get("back", 14))
    outfile = os.path.expanduser(config["DEFAULT"]["agenda_outfile"])

    if args.url:
        print("\n\n".join(dict.fromkeys(org_events(calendars, ahead, back))))
        return

    with open(outfile, "w") as fid:
        LOGGER.info("Writing calendars to: %s", outfile)
        fid.write("\n\n".join(dict.fromkeys(org_events(calendars, ahead, back))))
        fid.write("\n")


def write_addressbook(config, args):
    "Write Contacts to file"
    addresses = get_resource(config, "addressbooks", args.force)
    outfile = os.path.expanduser(config["DEFAULT"]["contacts_outfile"])

    with open(outfile, "w") as fid:
        LOGGER.info("Writing contacts to: %s", outfile)
        fid.write("\n\n".join(org_contacts(addresses)))
        fid.write("\n")


def get_config(filepath):
    "parse config file"

    config = configparser.ConfigParser()

    config_file = os.path.expanduser(filepath)
    LOGGER.info("Reading config from: %s", config_file)
    config.read([config_file])

    return config


def parse_arguments():
    "Parse CLI"
    parser = argparse.ArgumentParser(description="Translate CalDav Agenda to orgfile")
    parser.add_argument(
        "-f", "--force", help="Force Download of Caldav files", action="store_true"
    )
    parser.add_argument(
        "--contacts", action="store_true", help="Download carddav to org-contacts"
    )
    parser.add_argument("-r", "--url", help="force direct download from url no auth")
    parser.add_argument("-v", "--verbose", action="count", default=0)

    return parser.parse_args()


def main():
    "Transform Calendars"
    args = parse_arguments()

    log_level = logging.INFO if args.verbose > 0 else logging.WARNING
    LOGGER.setLevel(log_level)

    config = get_config("~/.calendars.conf")

    if args.contacts:
        write_addressbook(config, args)

    write_agenda(config, args)



# -*- coding: utf-8 -*-
r"""
Convert Caldav to org-contacts
==============================
"""
# Author: Óscar Nájera
# License: GPL-3
# Inspired https://gist.github.com/tmalsburg/9747104

import dateutil.parser
import vobject  # type: ignore
from org_agenda import org


def get_properties(contact):
    "Extract all contact elements as properties"
    ignore = ["VERSION", "PRODID", "FN", "NOTE", "CATEGORIES"]

    for prop in contact.getChildren():

        name = prop.name
        value = prop.value
        # Special treatment for some fields:
        if prop.name in ignore or prop.name.startswith("X-"):
            continue

        if prop.name == "N":
            value = "%s;%s;%s;%s;%s" % (
                prop.value.family,
                prop.value.given,
                prop.value.additional,
                prop.value.prefix,
                prop.value.suffix,
            )

        if prop.name == "ADR":
            value = (
                prop.value.street,
                prop.value.code + " " + prop.value.city,
                prop.value.region,
                prop.value.country,
                prop.value.extended,
                prop.value.box,
            )
            value = ", ".join([x for x in value if x.strip() != ""])
            name = "ADDRESS"

        if prop.name == "REV":
            value = dateutil.parser.parse(prop.value)
            value = value.strftime("[%Y-%m-%d %a %H:%M]")

        if prop.name == "TEL":
            name = "PHONE"

        # Collect type attributes:
        if attribs := ", ".join(prop.params.get("TYPE", [])):
            attribs = " (%s)" % attribs

        # Make sure that there are no newline chars left as that would
        # break org's property format:
        if isinstance(value, (list, tuple)):
            value = ", ".join(value)
        if value := value.replace("\n", ", "):
            yield name, value + attribs


class OrgContact(org.OrgEntry):
    "Contact representation in Org-mode"

    def __init__(self, contact):
        super().__init__(contact)
        self.property_parser = get_properties
        self.heading = contact.fn.value
        self.description = contact.getChildValue("note", "")
        self.dates = ""

    @property
    def tags(self):
        "Tags"
        return org.tags(self.entry.getChildValue("categories", []))


def org_contacts(addressbooks):
    "Iterate all addressbooks to generate contacts"
    for book in map(vobject.readComponents, addressbooks):
        for contact in book:
            yield str(OrgContact(contact))



# -*- coding: utf-8 -*-
r"""
Convert Caldav to org-contacts
==============================
"""
# Author: Óscar Nájera
# License: GPL-3
# Inspired https://gist.github.com/tmalsburg/9747104


def tags(tag_list):
    "Tags"

    if _tags := ":".join(tag_list):
        return f"  :{_tags}:"
    return ""


def string(node):
    "string representation of org entry"
    return (
        f"* {node.heading}{node.tags}\n"
        f"{node.properties}{node.dates}{node.description}"
    ).strip()


class OrgEntry:
    "Org-Mode basic entry"

    def __init__(self, entry):
        self.entry = entry
        self.property_parser = lambda x: []

    @property
    def properties(self):
        "Property box"
        props = "\n".join(
            ":%s: %s" % (k, v) for k, v in self.property_parser(self.entry)
        )
        return f""":PROPERTIES:\n{props}\n:END:\n""" if props else ""

    def __str__(self):
        return string(self)



# -*- coding: utf-8 -*-
r"""
Convert icalendar elements into org-mode elements
=================================================
"""
# Created: Thu Mar 19 12:48:59 2020
# Author: Óscar Nájera
# License: GPL-3

from datetime import datetime, timedelta
from dateutil import tz
from dateutil.rrule import rrulestr
from icalendar import Calendar  # type: ignore
from pytz import utc
from tzlocal import get_localzone  # type: ignore
from org_agenda import org

# inspiration from
# https://www.nylas.com/blog/calendar-events-rrules/


def orgdatetime(datestamp, time_zone, time=True):
    """Timezone aware datetime to YYYY-MM-DD DayofWeek HH:MM str in localtime.
    """
    hours = " %H:%M" if time else ""
    str_format = f"<%Y-%m-%d %a{hours}>"

    return datestamp.astimezone(time_zone).strftime(str_format)


def org_interval(start, duration, time_zone):
    "Write org interval"
    if duration.total_seconds() % 86400 == 0:
        datestr = "  {}".format(orgdatetime(start, time_zone, False))
        if duration.total_seconds() / 86400 > 1:
            datestr += "--{}".format(
                orgdatetime(start + duration - timedelta(seconds=1), time_zone, False)
            )
        return datestr + "\n"

    return "  {}--{}\n".format(
        orgdatetime(start, time_zone), orgdatetime(start + duration, time_zone),
    )


def put_tz(date_time):
    "Given date or datetime enforce transformation to localtime"
    if not hasattr(date_time, "hour"):
        return datetime(
            year=date_time.year,
            month=date_time.month,
            day=date_time.day,
            tzinfo=tz.tzlocal(),
        )
    return date_time.astimezone(tz.tzlocal())


def get_properties(event):
    "Extract relevant properties"
    properties = ("LOCATION", "UID")
    for prop in properties:
        if value := event.get(prop, "").strip():
            if prop == "UID":
                prop = "ID"
            yield prop, value

    if "RRULE" in event:
        yield "RRULE", rrule_cleanup(event["RRULE"])

    for comp in event.subcomponents:
        if comp.name == "VALARM":
            trigger = int(-1 * comp["TRIGGER"].dt.total_seconds() / 60)
            yield "APPT_WARNTIME", str(trigger)


def rrule_cleanup(rrule_conf):
    "Repetition rule needs to respect some constrains, return clean string"
    if until := rrule_conf.get("UNTIL"):
        if not hasattr(until[0], "hour"):
            until[0] = put_tz(until[0])
        rrule_conf["UNTIL"][0] = until[0].astimezone(tz.UTC)

    return rrule_conf.to_ical().decode("utf-8")


def rrule(org_event, exceptions=None):
    "Create event repetition rule"

    rule = rrulestr(
        rrule_cleanup(org_event.entry["RRULE"]),
        dtstart=org_event.dtstart,
        forceset=True,
    )

    exdates = org_event.entry.get("EXDATE", [])
    for dates in (exdates,) if not isinstance(exdates, list) else exdates:
        for date in dates.dts:
            rule.exdate(put_tz(date.dt))

    if exceptions:
        for date in exceptions:
            new_datetime = put_tz(date.dt).replace(
                hour=org_event.dtstart.hour, minute=org_event.dtstart.minute
            )
            rule.exdate(new_datetime)

    return rule


class OrgEvent(org.OrgEntry):
    """Documentation for OrgEntry"""

    def __init__(self, event):
        super().__init__(event)
        self.property_parser = get_properties
        self.heading = event["SUMMARY"]
        self.dtstart = put_tz(event["DTSTART"].dt)
        if "DTEND" in event:
            self.dtend = put_tz(event["DTEND"].dt)
            self.duration = self.dtend - self.dtstart
        else:
            self.duration = event["DURATION"].dt

        self.dates = ""

        self.description = (
            event["DESCRIPTION"].replace(" \n", "\n") if "DESCRIPTION" in event else ""
        )

    @property
    def tags(self):
        "Tags"

        _tags = self.entry.get("CATEGORIES", [])
        if not isinstance(_tags, list):
            _tags = [_tags]

        _tags = map(
            lambda x: x.to_ical().decode("utf-8").replace(" ", "-").replace(",", ":"),
            _tags,
        )

        if "RRULE" in self.entry or "RECURRENCE-ID" in self.entry:
            _tags = list(_tags) + ["rrule"]

        return org.tags(_tags)

    def date_block(self, start, end, exceptions=None):
        "Evaluate which active dates the event has"

        if "RRULE" in self.entry:
            rule = rrule(self, exceptions)
            self.dates = "".join(
                org_interval(event_start, self.duration, get_localzone())
                for event_start in rule.between(after=start, before=end)
            )

        elif self.dtstart < end and self.dtstart > start:
            self.dates = org_interval(self.dtstart, self.duration, get_localzone())

        return self.dates


def changev(evlist, start, end):
    "Clean repeating occurrences that have a specific event change"
    mods = [e.entry.get("RECURRENCE-ID") for e in evlist]

    if len(evlist) == 1 or None not in mods:
        return evlist

    base_ind = mods.index(None)
    mods.pop(base_ind)

    base = evlist.pop(base_ind)
    base.date_block(start, end, mods)

    return ([base] if base.dates else []) + evlist


def org_calendar(ical, start, end):
    "Return calendar time relevant calendar events"
    events = (OrgEvent(entry) for entry in ical.walk() if entry.name == "VEVENT")
    regular_events = {}
    for event in events:
        if event.date_block(start, end):
            regular_events.setdefault(event.entry.get("UID"), []).append(event)

    for evlist in regular_events.values():
        yield from map(str, changev(evlist, start, end))


def org_events(calendars, ahead, back):
    "Iterator of all events in calendars from [today-back;today+ahead]"

    now = datetime.now(utc)
    start = now - timedelta(back)
    end = now + timedelta(ahead)

    for ical in map(Calendar.from_ical, calendars):
        yield from org_calendar(ical, start, end)



from setuptools import setup

package = "org-mode-agenda"
version = "0.4"

with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name=package,
    version=version,
    author="Óscar Nájera",
    author_email="hi@oscarnajera.com",
    description="Import icalendar agendas to Orgmode",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://gitlab.com/Titan-C/org-mode-agenda",
    packages=["org_agenda"],
    entry_points={"console_scripts": ["org_agenda_sync = org_agenda:main"]},
    install_requires=["requests", "icalendar", "tzlocal", "vobject", "aiohttp"],
    license="GNU General Public License v3 or later (GPLv3+)",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
    python_requires=">=3.6",
)
